from copy import deepcopy
from io import BytesIO, BufferedReader
from typing import Dict, NamedTuple, Callable, Optional, Union

import pytest
import requests
import json
import inspect
import tempfile
from platform import system

from pytest_httpserver import HTTPServer, HeaderValueMatcher
from pytest import mark

from arthurai.client.http.arthur import ArthurHTTPClient
from arthurai.client.http.base import AbstractHTTPClient
from arthurai.client.http.requests import HTTPClient
from arthurai.client.client import ORG_ID_BODY
from arthurai import __version__ as arthur_version

# CONSTANTS
from tests import MockResponse

MULTIPART_HEADER_PREFIX = "multipart/form-data; boundary="

# FIXTURES
# to be supplied
ACCESS_KEY = "abc123"
SAMPLE_DATA = {'foo': "bar",
               'baz': 2}
ARRAY_DATA = [1, 2, 3]
SAMPLE_DATA_BYTES = bytes(json.dumps(SAMPLE_DATA), encoding='utf-8')
ARRAY_DATA_BYTES = bytes(json.dumps(ARRAY_DATA), encoding='utf-8')
NESTED_DATA = {'data': json.dumps(SAMPLE_DATA), 'other_data': ARRAY_DATA}
NESTED_DATA_BYTES = {'data': SAMPLE_DATA_BYTES, 'other_data': ARRAY_DATA_BYTES}
PARAMS = {'query': "param"}
SAMPLE_FILE_DATA = {
    "file.name": bytes("hello im mr data", encoding="utf-8"),
    "other.file": bytes("and im other data", encoding="utf-8"),
}

# to be returned
ORG_ID = "org123"
ORG_ID_RESPONSE = {ORG_ID_BODY: ORG_ID}

# headers
INITIAL_HEADERS = {
    "Authorization": ACCESS_KEY,
    "Accept": "application/json"
}
JSON_HEADERS = {
    **INITIAL_HEADERS, "Content-Type": "application/json"}
MULTIPART_HEADERS = {
    **INITIAL_HEADERS, "Content-Type": "multipart/form-data"
}

EXPECTED_USER_AGENT = f"arthur-sdk/{arthur_version} (system={system()}, org={ORG_ID})"


def given_arthur_http_client(httpserver: HTTPServer) -> ArthurHTTPClient:
    base_url = httpserver.url_for("")
    return ArthurHTTPClient(access_key=ACCESS_KEY, url=base_url, base_path='/api/v3', verify_ssl=True)


def given_requests_http_client(httpserver: HTTPServer) -> HTTPClient:
    base_url = httpserver.url_for("")
    headers = {
        'Accept': "application/json",
        'Authorization': ACCESS_KEY,
        'User-Agent': EXPECTED_USER_AGENT,
    }
    return HTTPClient(base_url=base_url, path_prefix="/api", default_headers=headers, verify_ssl=True)


"""
TIP:
The pytest_httpserver lib is generally great but a little annoying because if there's no match you just get back a 
generic "no matcher found" response. I suggest dropping a breakpoint in the httpserver.py/RequestMatcher.match()
method if you're trying to diagnose errors.
"""


class TestHTTPClientCase(NamedTuple):
    given_client: Callable[[HTTPServer], AbstractHTTPClient]
    user_agent: str
    api_prefix: str
    extra_requests_count: int
    multipart_requires_content_type: bool
    bytes_io_instead_of_file: bool


CASES = [
    pytest.param(
        TestHTTPClientCase(given_client=given_requests_http_client, user_agent=EXPECTED_USER_AGENT, api_prefix="/api",
                           extra_requests_count=0, multipart_requires_content_type=False,
                           bytes_io_instead_of_file=False), id="requests-http-client-file"),
    pytest.param(
        TestHTTPClientCase(given_client=given_requests_http_client, user_agent=EXPECTED_USER_AGENT, api_prefix="/api",
                           extra_requests_count=0, multipart_requires_content_type=False,
                           bytes_io_instead_of_file=True), id="requests-http-client-bytes-io")
]


@mark.parametrize("case", CASES)
def test_client_method_signatures_match(httpserver: HTTPServer, case: TestHTTPClientCase):
    allowed_method_name_differences = {"endpoint": "url"}
    abstract_methods = AbstractHTTPClient.__abstractmethods__
    target_class = inspect.signature(case.given_client).return_annotation
    for method_name in abstract_methods:
        abstract_signature = inspect.signature(AbstractHTTPClient.__getattribute__(AbstractHTTPClient, method_name))
        concrete_signature = inspect.signature(target_class.__getattribute__(target_class, method_name))
        concrete_signature_list = list(concrete_signature.parameters.items())
        i = 0
        for abstract_name, abstract_param in abstract_signature.parameters.items():
            if abstract_param.kind.name == "POSITIONAL_ONLY":
                concrete_name, concrete_param = concrete_signature_list[i]
            elif abstract_param.kind.name in ("KEYWORD_ONLY", "POSITIONAL_OR_KEYWORD"):
                if abstract_name in concrete_signature.parameters.keys():
                    concrete_name = abstract_name
                elif abstract_name in allowed_method_name_differences.keys() and allowed_method_name_differences[
                    abstract_name] in concrete_signature.parameters.keys():
                    concrete_name = allowed_method_name_differences[abstract_name]
                else:
                    raise AssertionError(f"abstract method '{method_name}' keyword parameter '{abstract_name}' not in "
                                         "concrete implementation")
                concrete_param = concrete_signature.parameters[concrete_name]
            elif abstract_param.kind.name in ("VAR_KEYWORD", "VAR_POSITIONAL"):
                continue
            else:
                raise ValueError(f"can't interpret paramter type {abstract_param.kind.name}")

            if abstract_param.default != inspect._empty:  # type: ignore
                assert concrete_param.default == abstract_param.default
            i += 1


@mark.parametrize("case", CASES)
def test_get(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    # allow the client to include an application/json Content-Type header even though it shouldn't
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(INITIAL_HEADERS, case.user_agent),
                                      query_string=PARAMS) \
        .respond_with_json({'status': "ok"})
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent),
                                      query_string=PARAMS) \
        .respond_with_json({'status': "ok"})

    response = client.get("/test", params=PARAMS, return_raw_response=True)

    assert response.status_code < 300
    assert len(httpserver.log) == 1 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_delete(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    # allow the client to include an application/json Content-Type header even though it shouldn't
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="DELETE",
                                      headers=add_user_agent_header(INITIAL_HEADERS, case.user_agent)) \
        .respond_with_json({'status': "ok"})
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="DELETE",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent)) \
        .respond_with_json({'status': "ok"})

    response = client.delete("/test", return_raw_response=True)

    assert response.status_code < 300
    assert len(httpserver.log) == 1 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_json_post(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    data_copy = deepcopy(SAMPLE_DATA)
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="POST", json=SAMPLE_DATA,
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent)) \
        .respond_with_json({'status': "ok"})

    response = client.post("/test", json=SAMPLE_DATA, return_raw_response=True)

    assert response.status_code < 300
    assert len(httpserver.log) == 1 + case.extra_requests_count
    # inputs not modified
    assert data_copy == SAMPLE_DATA


@mark.parametrize("case", CASES)
def test_multipart_post(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    sample_files: Optional[Union[Dict[str, BytesIO], Dict[str, tempfile._TemporaryFileWrapper]]] = None
    if case.bytes_io_instead_of_file:
        sample_files = sample_file_data_to_bytes_io_map(SAMPLE_FILE_DATA)
    else:
        sample_files = sample_file_data_to_files_map(SAMPLE_FILE_DATA)

    # header matchers, kinda weird API but this does the trick
    multipart_header_matcher = HeaderValueMatcher(matchers={
        **{header: HeaderValueMatcher.default_header_value_matcher for header in INITIAL_HEADERS.keys()},
        'User-Agent': HeaderValueMatcher.default_header_value_matcher,
        'Content-Type': match_multipart_headers
    })

    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="POST",
                                      header_value_matcher=multipart_header_matcher,
                                      headers=add_user_agent_header(MULTIPART_HEADERS, case.user_agent)) \
        .respond_with_json({'status': "ok"})

    # our client makes you specify multipart in the headers yourself, otherwise it wont send the files you supply
    base_headers = {'Content-Type': "multipart/form-data"} if case.multipart_requires_content_type else {}
    response = client.post("/test", json=None, files=sample_files,
                           headers=add_user_agent_header(base_headers, case.user_agent),
                           return_raw_response=True)

    assert response.status_code < 300
    request = httpserver.log[-1][0]
    assert_multipart_expected(request, files_map=sample_files, data_map={})
    assert len(httpserver.log) == 1 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_multipart_post_retry(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    sample_files: Optional[Union[Dict[str, BytesIO], Dict[str, tempfile._TemporaryFileWrapper]]] = None
    if case.bytes_io_instead_of_file:
        sample_files = sample_file_data_to_bytes_io_map(SAMPLE_FILE_DATA)
    else:
        sample_files = sample_file_data_to_files_map(SAMPLE_FILE_DATA)

    # header matchers, kinda weird API but this does the trick
    multipart_header_matcher = HeaderValueMatcher(matchers={
        **{header: HeaderValueMatcher.default_header_value_matcher for header in INITIAL_HEADERS.keys()},
        'User-Agent': HeaderValueMatcher.default_header_value_matcher,
        'Content-Type': match_multipart_headers
    })

    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="POST",
                                      header_value_matcher=multipart_header_matcher,
                                      headers=add_user_agent_header(MULTIPART_HEADERS, case.user_agent)) \
        .respond_with_json({'errors': "no bad"}, status=500)
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="POST",
                                      header_value_matcher=multipart_header_matcher,
                                      headers=add_user_agent_header(MULTIPART_HEADERS, case.user_agent)) \
        .respond_with_json({'message': "ok"})

    # our client makes you specify multipart in the headers yourself, otherwise it wont send the files you supply
    base_headers = {'Content-Type': "multipart/form-data"} if case.multipart_requires_content_type else {}
    response = client.post("/test", json=None, files=sample_files,
                           headers=add_user_agent_header(base_headers, case.user_agent), retries=1,
                           return_raw_response=True)

    assert response.status_code < 300
    request = httpserver.log[-1][0]
    assert_multipart_expected(request, files_map=sample_files, data_map={})
    assert len(httpserver.log) == 2 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_multipart_post_mixed_retry(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    sample_files: Optional[Union[Dict[str, BytesIO], Dict[str, tempfile._TemporaryFileWrapper]]] = None
    if case.bytes_io_instead_of_file:
        sample_files = sample_file_data_to_bytes_io_map(SAMPLE_FILE_DATA)
    else:
        sample_files = sample_file_data_to_files_map(SAMPLE_FILE_DATA)

    # header matchers, kinda weird API but this does the trick
    multipart_header_matcher = HeaderValueMatcher(matchers={
        **{header: HeaderValueMatcher.default_header_value_matcher for header in INITIAL_HEADERS.keys()},
        'User-Agent': HeaderValueMatcher.default_header_value_matcher,
        'Content-Type': match_multipart_headers
    })

    data_copy = deepcopy(NESTED_DATA)

    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="POST",
                                      header_value_matcher=multipart_header_matcher,
                                      headers=add_user_agent_header(MULTIPART_HEADERS, case.user_agent)) \
        .respond_with_json({'errors': "no bad"}, status=500)
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="POST",
                                      header_value_matcher=multipart_header_matcher,
                                      headers=add_user_agent_header(MULTIPART_HEADERS, case.user_agent)) \
        .respond_with_json({'message': "ok"})

    # our client makes you specify multipart in the headers yourself, otherwise it wont send the files you supply
    base_headers = {'Content-Type': "multipart/form-data"} if case.multipart_requires_content_type else {}
    response = client.post("/test", json=NESTED_DATA, files=sample_files,
                           headers=add_user_agent_header(base_headers, case.user_agent), retries=1,
                           return_raw_response=True)

    assert response.status_code < 300
    request = httpserver.log[-1][0]
    assert_multipart_expected(request, files_map=sample_files, data_map=NESTED_DATA_BYTES)
    assert len(httpserver.log) == 2 + case.extra_requests_count
    # inputs not modified
    assert data_copy == NESTED_DATA


@mark.parametrize("case", CASES)
def test_multipart_post_nofiles(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    data_copy = deepcopy(NESTED_DATA)

    # header matchers, kinda weird API but this does the trick
    multipart_header_matcher = HeaderValueMatcher(matchers={
        **{header: HeaderValueMatcher.default_header_value_matcher for header in INITIAL_HEADERS.keys()},
        'User-Agent': HeaderValueMatcher.default_header_value_matcher,
        'Content-Type': match_multipart_headers
    })

    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="POST",
                                      header_value_matcher=multipart_header_matcher, headers=MULTIPART_HEADERS) \
        .respond_with_json({'status': "ok"})

    response = client.post("/test", json=NESTED_DATA,
                           headers=add_user_agent_header({'Content-Type': "multipart/form-data"}, case.user_agent),
                           return_raw_response=True)

    assert response.status_code < 300
    request = httpserver.log[-1][0]
    assert_multipart_expected(request, data_map=NESTED_DATA_BYTES)
    assert len(httpserver.log) == 1 + case.extra_requests_count
    # inputs not modified
    assert data_copy == NESTED_DATA


@mark.parametrize("case", CASES)
def test_json_patch(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    data_copy = deepcopy(SAMPLE_DATA)
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="PATCH", json=SAMPLE_DATA,
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent)) \
        .respond_with_json({'status': "ok"})

    response = client.patch("/test", json=SAMPLE_DATA, return_raw_response=True)

    assert response.status_code < 300
    assert len(httpserver.log) == 1 + case.extra_requests_count
    # inputs not modified
    assert data_copy == SAMPLE_DATA


@mark.parametrize("case", CASES)
def test_multipart_patch_mixed(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    sample_files: Optional[Union[Dict[str, BytesIO], Dict[str, tempfile._TemporaryFileWrapper]]] = None
    if case.bytes_io_instead_of_file:
        sample_files = sample_file_data_to_bytes_io_map(SAMPLE_FILE_DATA)
    else:
        sample_files = sample_file_data_to_files_map(SAMPLE_FILE_DATA)

    # header matchers, kinda weird API but this does the trick
    multipart_header_matcher = HeaderValueMatcher(matchers={
        **{header: HeaderValueMatcher.default_header_value_matcher for header in INITIAL_HEADERS.keys()},
        'Content-Type': match_multipart_headers
    })

    data_copy = deepcopy(NESTED_DATA)
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="PATCH",
                                      header_value_matcher=multipart_header_matcher, headers=MULTIPART_HEADERS) \
        .respond_with_json({'status': "ok"})

    response = client.patch("/test", json=NESTED_DATA, files=sample_files,
                            headers=add_user_agent_header({'Content-Type': "multipart/form-data"}, case.user_agent),
                            return_raw_response=True)

    assert response.status_code < 300
    request = httpserver.log[-1][0]
    assert_multipart_expected(request, files_map=sample_files, data_map=NESTED_DATA_BYTES)
    assert len(httpserver.log) == 1 + case.extra_requests_count
    # inputs not modified
    assert data_copy == NESTED_DATA


@mark.parametrize("case", CASES)
def test_put(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    data_copy = deepcopy(SAMPLE_DATA)
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="PUT", json=SAMPLE_DATA,
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent)) \
        .respond_with_json({'status': "ok"})

    response = client.put("/test", json=SAMPLE_DATA, return_raw_response=True)

    assert response.status_code < 300
    # one for get org and one for current
    assert len(httpserver.log) == 1 + case.extra_requests_count
    # inputs not modified
    assert data_copy == SAMPLE_DATA


@mark.parametrize("case", CASES)
def test_retries_validation(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    # headers
    def allow_none_match(actual: str, expected: str) -> bool:
        return (actual == expected) or actual is None

    allow_none_matcher = HeaderValueMatcher(matchers={
        **{header: HeaderValueMatcher.default_header_value_matcher for header in INITIAL_HEADERS.keys()},
        'User-Agent': HeaderValueMatcher.default_header_value_matcher,
        'Content-Type': allow_none_match
    })

    # 400 response
    bad_request_response = MockResponse("bad request", 400)
    internal_error_response = MockResponse("internal server error", 500)

    if 'validation_response_code' in inspect.signature(client.send).parameters.keys():
        additional_args = {'validation_response_code': 200}
    else:
        additional_args = {}

    # two failures then one success
    httpserver.expect_ordered_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent),
                                      header_value_matcher=allow_none_matcher) \
        .respond_with_response(bad_request_response)
    httpserver.expect_ordered_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent),
                                      header_value_matcher=allow_none_matcher) \
        .respond_with_response(internal_error_response)
    httpserver.expect_ordered_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent),
                                      header_value_matcher=allow_none_matcher) \
        .respond_with_json({'status': "ok"})

    response = client.send("/test", method="GET", return_raw_response=True, retries=2, **additional_args)

    assert response.status_code < 300
    assert len(httpserver.log) == 3 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_retries_validation_failure(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    # headers
    def allow_none_match(actual: str, expected: str) -> bool:
        return (actual == expected) or actual is None

    allow_none_matcher = HeaderValueMatcher(matchers={
        **{header: HeaderValueMatcher.default_header_value_matcher for header in INITIAL_HEADERS.keys()},
        'User-Agent': HeaderValueMatcher.default_header_value_matcher,
        'Content-Type': allow_none_match
    })

    # 400 response
    bad_request_response = MockResponse("bad request", 400)
    internal_error_response = MockResponse("internal server error", 500)

    if 'validation_response_code' in inspect.signature(client.send).parameters.keys():
        additional_args = {'validation_response_code': 200}
    else:
        additional_args = {}

    # two failures then one success
    httpserver.expect_ordered_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent),
                                      header_value_matcher=allow_none_matcher) \
        .respond_with_response(bad_request_response)
    httpserver.expect_ordered_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent),
                                      header_value_matcher=allow_none_matcher) \
        .respond_with_response(internal_error_response)

    exception = None
    response = None
    try:
        response = client.send("/test", method="GET", return_raw_response=True, retries=1, **additional_args)
    except Exception as e:
        exception = e

    assert exception is not None or (response is not None and response.status_code == 500)
    assert len(httpserver.log) == 2 + case.extra_requests_count


def add_user_agent_header(original_headers, user_agent):
    return {
        **original_headers,
        "User-Agent": user_agent
    }


def given_orgid_returned(httpserver: HTTPServer) -> None:
    # allow the client to include an application/json Content-Type header even though it shouldn't
    headers = {**INITIAL_HEADERS, "Content-Type": "application/json"}
    user_agent = f"python-requests/{requests.__version__}"
    httpserver.expect_oneshot_request("/api/v3/users/me", method="GET",
                                      headers=add_user_agent_header(headers, user_agent)) \
        .respond_with_json(ORG_ID_RESPONSE)
    httpserver.expect_oneshot_request("/api/v3/users/me", method="GET",
                                      headers=add_user_agent_header(INITIAL_HEADERS, user_agent)) \
        .respond_with_json(ORG_ID_RESPONSE)


def match_multipart_headers(actual: str, expected: str) -> bool:
    return actual.startswith("multipart/form-data; boundary=")


def sample_file_data_to_files_map(sample_data: Dict[str, bytes]) -> Dict[str, tempfile._TemporaryFileWrapper]:
    file_map = {}
    for fname, content in sample_data.items():
        temp_file = tempfile.NamedTemporaryFile()
        with open(temp_file.name, 'wb') as f:
            f.write(content)
        file_map[fname] = temp_file
    return file_map


def sample_file_data_to_bytes_io_map(sample_data: Dict[str, bytes]) -> Dict[str, BytesIO]:
    return {
        fname: BytesIO(content) for fname, content in sample_data.items()
    }


def assert_multipart_expected(request: requests.Request, files_map: Optional[Union[Dict[str, BytesIO], Dict[str, tempfile._TemporaryFileWrapper]]] = None,
                              data_map: Optional[Dict[str, bytes]] = None) -> None:
    if files_map is None:
        replacement_files_map: Dict[str, BytesIO] = {}
        files_map = replacement_files_map
    if data_map is None:
        data_map = {}
    boundary = request.headers['Content-Type'][len(MULTIPART_HEADER_PREFIX):]
    prefix = bytes(f"--{boundary}", encoding="utf-8")

    data = request.data
    assert data is not None
    assert isinstance(data, bytes)
    assert data.startswith(prefix)

    blocks = data.split(prefix)

    # build map from expected content to count of instances
    block_counts_map = {}
    for filename, content in files_map.items():
        content.seek(0)
        file_contents = content.file.read() if hasattr(content, 'file') else content.read()
        passed_filename = content.name if hasattr(content, 'name') else filename
        expected_block = bytes(
            f'\r\nContent-Disposition: form-data; name="{filename}"; filename="{passed_filename}"\r\n\r\n',
            encoding='utf-8') + file_contents + bytes("\r\n", encoding='utf-8')
        block_counts_map[expected_block] = 0
    for keyname, data in data_map.items():
        expected_block = bytes(f'\r\nContent-Disposition: form-data; name="{keyname}"\r\n\r\n',
                               encoding='utf-8') + data + bytes("\r\n", encoding='utf-8')
        block_counts_map[expected_block] = 0

    # first block should be empty per startswith() assertion above, iterate through content blocks
    # go through each subsequent block, it should be in our map
    for block in blocks[1:-1]:
        assert block in block_counts_map.keys()
        block_counts_map[block] += 1

    # ensure each block was seen exactly once
    for count in block_counts_map.values():
        assert count == 1

    # validate ends with expected suffix
    assert blocks[-1] == bytes("--\r\n", encoding='utf-8')



import pytest
import pandas as pd
from arthurai import ArthurAttribute, ArthurModel
from arthurai.common.constants import InputType, OutputType, ValueType, Stage
from arthurai.core.attributes import AttributeCategory
from tests.fixtures.models.multiclass import DOG_PRED, CAT_PRED, GROUND_TRUTH, UNLINKED_CAT_PRED, UNLINKED_DOG_PRED


INPUT_TEXT = ["hello there", "hi", "what's up", "yo", "cool", "cool beans", "this", "should", "be", "unique", "yay"]
NLP_DATAFRAME = pd.DataFrame({"input_text": INPUT_TEXT,
                              "pred": [90, 80, 50, 31, 26, 18, 90, 80, 50, 31, 26],
                              "gt": [95, 81, 42, 30, 30, 27, 95, 81, 42, 30, 30]})


DOG_CLASS_STRING = "dog class"
CAT_CLASS_STRING = "cat_class"
NLP_CLASSIFICATION_DATAFRAME = pd.DataFrame({"input_text": INPUT_TEXT,
                                             DOG_PRED: [0.6, 0.8, 0.2, 0.45, 0.94, 0.12, 0.6, 0.8, 0.2, 0.45, 0.94],
                                             CAT_PRED: [0.4, 0.2, 0.8, 0.65, 0.06, 0.88, 0.4, 0.2, 0.8, 0.65, 0.06],
                                             GROUND_TRUTH: [DOG_CLASS_STRING, CAT_CLASS_STRING, DOG_CLASS_STRING, CAT_CLASS_STRING,
                                                            CAT_CLASS_STRING, DOG_CLASS_STRING, DOG_CLASS_STRING, CAT_CLASS_STRING,
                                                            DOG_CLASS_STRING, CAT_CLASS_STRING,CAT_CLASS_STRING]})

INPUT_TEXT_ATTR = ArthurAttribute(name='input_text',
                                  value_type=ValueType.Unstructured_Text,
                                  stage=Stage.ModelPipelineInput,
                                  # TODO: this seems like it should be false,
                                  #  but check that this field isn't used somewhere else where it needs to be true
                                  categorical=True,
                                  position=0,
                                  is_unique=True)

OUTPUT_TEXT_ATTR = ArthurAttribute(name='output_text',
                                   value_type=ValueType.Unstructured_Text,
                                   stage=Stage.PredictedValue,
                                   categorical=True,
                                   position=0,
                                   is_unique=False)

OUTPUT_TEXT_ATTR_LINKED = ArthurAttribute(name='output_text',
                                          value_type=ValueType.Unstructured_Text,
                                          stage=Stage.PredictedValue,
                                          categorical=True,
                                          position=0,
                                          is_unique=False,
                                          token_attribute_link="output_tokens")

OUTPUT_TOKENS_ATTR = ArthurAttribute(name='output_tokens',
                                     value_type=ValueType.Tokens,
                                     stage=Stage.PredictedValue,
                                     categorical=False,
                                     position=1,
                                     is_unique=False,
                                     token_attribute_link="output_text")

OUTPUT_PROBS_ATTR = ArthurAttribute(name='output_probs',
                                    value_type=ValueType.TokenLikelihoods,
                                    stage=Stage.PredictedValue,
                                    categorical=False,
                                    position=2,
                                    is_unique=False)

GT_STRING_ATTR = ArthurAttribute(name=GROUND_TRUTH,
                                 value_type=ValueType.String,
                                 stage=Stage.GroundTruthClass,
                                 categorical=True,
                                 categories=[AttributeCategory(value=DOG_CLASS_STRING),
                                             AttributeCategory(value=CAT_CLASS_STRING)],
                                 position=0)

GT_TEXT_ATTR = ArthurAttribute(name='gt_text',
                               value_type=ValueType.Unstructured_Text,
                               stage=Stage.GroundTruth,
                               categorical=True,
                               position=0,
                               is_unique=True,
                               token_attribute_link="gt_tokens")

GT_TOKENS_ATTR = ArthurAttribute(name='gt_tokens',
                                 value_type=ValueType.Tokens,
                                 stage=Stage.GroundTruth,
                                 categorical=False,
                                 position=1,
                                 is_unique=False,
                                 token_attribute_link="gt_text")

CLASSIFICATION_MODEL_ATTRIBUTES = [UNLINKED_CAT_PRED, UNLINKED_DOG_PRED, INPUT_TEXT_ATTR, GT_STRING_ATTR]
SEQUENCE_MODEL_ATTRIBUTES = [INPUT_TEXT_ATTR, OUTPUT_TEXT_ATTR]
SEQUENCE_MODEL_WITH_PROBS_ATTRIBUTES = [INPUT_TEXT_ATTR, OUTPUT_TEXT_ATTR_LINKED, OUTPUT_TOKENS_ATTR, OUTPUT_PROBS_ATTR]
SEQUENCE_MODEL_WITH_GT_ATTRIBUTES = [INPUT_TEXT_ATTR, OUTPUT_TEXT_ATTR, GT_TEXT_ATTR, GT_TOKENS_ATTR]

UNINITIALIZED_MODEL_DATA = {"partner_model_id": "test",
                            "client": None,
                            "input_type": InputType.NLP}


@pytest.fixture
def initial_nlp_regression_model():
    return ArthurModel(
        **UNINITIALIZED_MODEL_DATA,
        output_type=OutputType.Regression
    )


@pytest.fixture
def initial_nlp_classification_model():
    return ArthurModel(
        **UNINITIALIZED_MODEL_DATA,
        output_type=OutputType.Multiclass
    )


@pytest.fixture
def initial_nlp_sequence_model():
    return ArthurModel(
        **UNINITIALIZED_MODEL_DATA,
        output_type=OutputType.TokenSequence
    )



from typing import Dict, List

import arthurai
import numpy as np
from arthurai.common.constants import Stage, OutputType, ValueType

def check_has_bias_attrs(model):
    """
    Returns True if and only if the model has any attributes being monitored for bias by Arthur
    """
    for attr in model.attributes:
        if attr.monitor_for_bias:
            return True 
    return False

def check_attr_is_bias(model, attr_name: str):
    """
    Returns True if and only if the model has an attribute with the given name `attr_name` being monitored for bias by
    Arthur
    """
    attr = model.get_attribute(attr_name)
    return attr.monitor_for_bias

def get_positive_predicted_class(model):
    """
    Checks if model is a binary classifier. Returns False if multiclass, otherwise returns the name of the positive
    predicted attribute
    """
    if model.output_type != OutputType.Multiclass:
        return("Not a classifier model.")
    predicted_value_attributes = model.get_attributes(stage=Stage.PredictedValue)
    if len(predicted_value_attributes) != 2:
        return ("Not a binary classifier.")
    for attr in predicted_value_attributes:
        if attr.is_positive_predicted_attribute:
            return attr.name


"""
Token Sequence Output Type Related Helpers
"""

EPSILON = 1e-3


def tensors_to_arthur_inference(log_tensor: np.array, id_to_vocab: Dict[int, str], num_probs: int
                                ) -> List[Dict[str, float]]:
    """
    Convert an array of log probabilities with shape (sequence length, vocab size) to an Arthur Token Likelihoods
    formatted object.

    :param log_tensor: array of log probabilities for a generated sequence
    :param id_to_vocab: mapping from token ids to vocab strings
    :param num_probs: the number of tokens at each index in the sequence to include (up to 5)
    """
    if num_probs > 5:
        raise ValueError("num_probs must be <= 5")

    # TODO: this function should probably be batch
    if log_tensor.ndim != 2:
        raise ValueError("expected token probabilities to be two-dimensional")

    arthur_likelihoods = []
    probabilities = np.exp(log_tensor)
    invalid_probs = np.abs(np.sum(probabilities, axis=1) - 1) >= EPSILON
    if np.sum(invalid_probs) != 0:
        raise ValueError("all token probabilities must sum to 1")

    # get num_probs argmax
    top_indices = np.argpartition(probabilities, -num_probs, axis=1)[:, -num_probs:]
    top_values = probabilities[np.vstack(np.arange(len(probabilities))), top_indices]
    for v, i in zip(top_values, top_indices):
        arthur_likelihoods.append({id_to_vocab[index]: value.item() for value, index in zip(v, i)})
    return arthur_likelihoods


def _text_attr(name: str, stage: Stage):
    return arthurai.ArthurAttribute(
        name=name,
        stage=stage,
        position=0,
        categorical=True,
        value_type=ValueType.Unstructured_Text,
        is_unique=False if stage == Stage.PredictedValue else True
    )


def _tokens_attr(name: str, stage: Stage, position: int = 0):
    return arthurai.ArthurAttribute(
        name=name,
        stage=stage,
        value_type=ValueType.Tokens,
        categorical=False,
        is_unique=False,
        position=position
    )


def _token_likelihoods_attr(name: str, position: int = 0):
    return arthurai.ArthurAttribute(
        name=name,
        stage=Stage.PredictedValue,
        value_type=ValueType.TokenLikelihoods,
        categorical=False,
        is_unique=False,
        position=position
    )


def _image_attr(name: str):
    return arthurai.ArthurAttribute(
        name=name,
        stage=Stage.ModelPipelineInput,
        value_type=ValueType.Image,
        categorical=False,
        position=0
    )



from datetime import datetime
from dateutil.parser import parse, ParserError
import logging
import numpy as np
import pandas as pd
from pandas import Series
from pandas._libs.parsers import is_float_dtype
import re
from typing import Any

from arthurai import ArthurAttribute
from arthurai.common.constants import ValueType
from arthurai.common.exceptions import UnexpectedValueError, UnexpectedTypeError, UserTypeError, UserValueError
from arthurai.core.util import is_str_like, is_float_like, is_int_like, NumpyEncoder, is_list_like, is_date_like, \
    is_bool_like
from arthurai import util as arthur_util

logger = logging.getLogger(__name__)


def validate_series_data_type(series: Series, attr: ArthurAttribute) -> None:
    """ Validates if the series data type matches the attribute's Arthur value type

    Raises a UserTypeError if there is a data type mismatch where the series data type is not valid for the
    passed attribute's value type. Logs a warning if the series data type cannot be inferred because of None objects
    or empty lists or dicts.

    :param series: the Series to infer the data type from
    :param attr: the ArthurAttribute that the Series corresponds to
    :raises: UnexpectedValueError: A parameter is None.
    :raises: UserTypeError: Found data type mismatch between series and attr.
    """

    if series is None or attr is None:
        raise UnexpectedValueError("Parameter was None.")

    # standardize series
    series = series.replace([np.inf, -np.inf], np.nan)
    nans_present = series.isnull().values.any()
    series = series.dropna()
    if len(series) != 0:
        if is_float_dtype(series) and nans_present and (attr.value_type == ValueType.Integer or
                                                        attr.value_type == ValueType.Float):
            if len(series) > 0 and np.array_equal(series, series.astype(int)):  # int col with nans received as floats
                series = series.astype(pd.Int64Dtype())

        # validate data type mismatch
        sample_datum = series.iloc[0]
        ensure_obj_matches_attr_value_type(sample_datum, attr.value_type, attr.name)
    else:
        logger.warning(f"The data type of column {attr.name} cannot be validated because the column only has "
                       f"None objects.")


def obj_value_type_mismatch_err(attr_name: str, expected: str, obj_type: str) -> UserTypeError:
    """ Returns UserTypeError with data type mismatch message
    :param: attr_name: Name of attribute whose value type has been mismatched with a column's data type
    :param: expected: String representing expected data type for the attribute
    :param: obj_type: String representing the object type that didn't match expected
    :raises: UnexpectedTypeError: Parameter was unexpectedly a None object.
    :returns: UserTypeError with mismatch message.
    """
    if attr_name is None or expected is None or obj_type is None:
        raise UnexpectedTypeError("Cannot pass None parameters.")
    return UserTypeError(f"Expected {expected} data type for attribute {attr_name} but got object of type {obj_type}. "
                        f"Please correct the data type mismatch by changing either the value_type field of the "
                        f"attribute or the data type of the column.")


def ensure_obj_matches_attr_value_type(obj: Any, value_type: ValueType, source: str) -> None:
    """ Validates the object data type matches the Arthur value type

    Raises a UserTypeError if the object data type does not match any valid data type corresponding to
    the Arthur Value Type of the passed attribute.

    :param obj: Object whose data type should be compared
    :param value_type: ValueType to validate against obj
    :param source: String representation of source of data for clear error messages
    :raises: UserTypeError when obj's data type does not match attr's value_type Value Type.
    """
    if source is None or obj is None or value_type is None:
        raise UnexpectedTypeError("Cannot pass None parameters.")

    if value_type == ValueType.Float:
        if not is_float_like(obj):
            raise obj_value_type_mismatch_err(source, "float", str(type(obj)))
    elif value_type == ValueType.Integer:
        if not is_int_like(obj):
            raise obj_value_type_mismatch_err(source, "int", str(type(obj)))
    elif value_type == ValueType.String or value_type == ValueType.Image or \
            value_type == ValueType.Unstructured_Text:
        if not is_str_like(obj):
            raise obj_value_type_mismatch_err(source, "str", str(type(obj)))
        elif value_type == ValueType.String:  # validate string obj doesn't parse to a valid timestamp
            try:
                parse(obj)
            except (ParserError, OverflowError):  # string attribute is not a timestamp
                pass
            else:
                raise UserTypeError(f"Column {source} parses to valid timestamps but its attribute value_type is "
                                    f"{ValueType.String}, not {ValueType.Timestamp}. Please change the attribute "
                                    f"value_type and ensure the column consists of timezone aware Datetime objects.")
    elif value_type == ValueType.Boolean:
        if not is_bool_like(obj):
            raise obj_value_type_mismatch_err(source, "bool", str(type(obj)))
    elif value_type == ValueType.Timestamp:
        if not is_date_like(obj):
            raise obj_value_type_mismatch_err(source, "datetime or pd.Timestamp", str(type(obj)))
        else:
            datetime_obj = NumpyEncoder.convert_value(obj)
            arthur_util.format_timestamp(datetime_obj, col_name=source)
    elif value_type == ValueType.Tokens:  # expected type is list[str]
        list_obj = get_first_elem_if_valid_list(obj, source, "list[str]")  # validates list
        if list_obj is not None and not is_str_like(list_obj):  # validates list element
            raise obj_value_type_mismatch_err(source, "list[str]", f"list[{str(type(list_obj))}]")
    elif value_type == ValueType.TokenLikelihoods:
        validate_token_likelihoods_type(obj, source)
    elif value_type == ValueType.BoundingBox:  # expected type is list[list[Union[float, int]]]
        list_obj = get_first_elem_if_valid_list(obj, source, "list[list[Union[int, float]]]")
        if list_obj is not None:
            if not is_list_like(list_obj):
                raise obj_value_type_mismatch_err(source, "list[list[Union[int, float]]]",
                                                   f"list[{str(type(list_obj))}]")
            elif len(list_obj) != 6 and not len(list_obj) == 0:  # warning logged in _get_first_elem_if_valid_list if length is 0
                    raise UserValueError(f"Bounding box column {source} should have six values in the nested "
                                         f"list but an entry in your column had {len(list_obj)} values. For a full "
                                         f"description of Arthur's bounding box format see: "
                                         f"https://docs.arthur.ai/docs/cv-onboarding#formatting-bounding-boxes.")
            else:  # validate inner list
                nested_obj = get_first_elem_if_valid_list(list_obj, source,
                                                          "list[list[Union[int, float]]]", "list")
                if nested_obj is not None and not (is_float_like(nested_obj) or is_int_like(nested_obj)):
                    raise obj_value_type_mismatch_err(source, "list[list[Union[int, float]]]",
                                                       f"list[list[{str(type(nested_obj))}]]")



def get_first_elem_if_valid_list(obj: Any, source: str, expected: str, outer_type: str = "") -> Any:
    """ Validates obj is list-like and non-empty and returns the first element.

    Raises a UserTypeError if obj is not list-like (is not a numpy array or python list). Logs a warning if the
    length of the list is 0 or the list has a None object.

    :param: obj: Object whose data type is being checked.
    :param: attr_name: String source of the object for clear error messages.
    :param: expected: String of expected type of the object for clear error messages.
    :param: outer_type: String representing the data type of the object obj was nested in for clear error messages. Defaults to empty string.
    :raises: UserTypeError if obj is not list-like.
    :returns: first element if obj is a list, None otherwise
    """
    if obj is None or source is None or expected is None or outer_type is None:
        raise UnexpectedValueError("Parameters cannot be None.")

    if not is_list_like(obj):  # validate list type
        if outer_type != "":
            raise obj_value_type_mismatch_err(source, expected, outer_type + "[" + str(type(obj)) + "]")
        else:
            raise obj_value_type_mismatch_err(source, expected, str(type(obj)))
    list_obj = NumpyEncoder.convert_value(obj)
    if outer_type != "":
        outer_type = " nested"
    if len(list_obj) == 0:
        logger.warning(f"The length of the first{outer_type} list in {source} was 0. If you are using lists to "
                       f"represent missing values instead of None objects, please change your empty lists to None "
                       f"objects. The type of the list cannot be inferred from a zero-length list, so the data type "
                       f"of {source} cannot be validated.")
        return None
    elif obj[0] is None:
        logger.warning(f"The first element in the first{outer_type} list in {source} was None. The type of the list cannot be "
                       f"inferred from a None object, so the data type of {source} cannot be validated.")
        return None
    else:
        return list_obj[0]


def validate_token_likelihoods_type(obj: Any, source: str) -> None:
    """ Validates obj data type matches TokenLikelihoods Value Type

    Raises a UserTypeError if the object's data type does not match any valid data type corresponding to the
    TokenLikelihoods Value Type. Logs a warning if any of the nested types—list, dict, or dict keys—are length 0
    or None objects. The valid data type for TokenLikelihoods is list[dict[str, Union[int, float]]].
    :param obj: Object to validate.
    :param source: String representation of the source of obj for clear error messages.
    :raises: UserTypeError when obj's data type does not match a TokenLikelihoods Value Type data type.
    """
    if obj is None or source is None:
        raise UnexpectedTypeError("Parameters cannot be None.")
    list_obj = get_first_elem_if_valid_list(obj, source, "list[dict[str, Union[int, float]]]")
    if list_obj is not None:  # validate list element type is dict
        if not isinstance(list_obj, dict):
            raise obj_value_type_mismatch_err(source, "list[dict[str, Union[int, float]]]",
                                               f"list[{str(type(list_obj))}]")
        if len(list_obj) == 0:
            logger.warning(f"The first dict in the first list in {source} was empty. The data type "
                           f"cannot be validated for this column because no type can be inferred for the dictionary "
                           f"keys or values.")
        else:  # validate key and value types of dict
            key, val = next(iter(list_obj.items()))
            if key is None or val is None:
                logger.warning(f"A key or value in the first dict in the first list in {source} was None. The data "
                               f"type cannot be validated for this column because a type can't be inferred from a "
                               f"None object.")
            elif not is_str_like(key):
                raise obj_value_type_mismatch_err(source, "list[dict[str, Union[int, float]]]",
                                                   f"list[dict[{str(type(key))}, {str(type(val))}]]")
            elif not (is_float_like(val) or is_int_like(val)):
                raise obj_value_type_mismatch_err(source, "list[dict[str, Union[int, float]]]",
                                                   f"list[dict[str, {str(type(val))}]]")

def validate_attr_names(col_names: pd.Index) -> None:
    """ Validates attribute names are correctly formatted

    Validates attribute names are made of alphanumeric characters and underscores and don't start with a number.

    :param: col_names: pd.Index of attribute names to be checked
    :raises: UserValueError: At least one attribute name is not correctly formatted
    """
    attr_name_regex = re.compile(r'^[A-Za-z_][A-Za-z0-9_]+$')
    misformatted_cols = []
    for col in col_names:
        if not attr_name_regex.match(col):
            misformatted_cols.append(col)
    if len(misformatted_cols) != 0:
        raise UserValueError(f"The following column names are misformatted: {str(misformatted_cols)[1:-1]}. Please "
                             f"ensure each name consists of only alphanumeric characters and underscores and "
                             f"doesn't start with a number.")
